import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import numpy as np
import random
from six import iteritems
import os

from .residuals import res_warp_img, res_warp_res, combine_residuals
from .model import Model

class ScaleNet(Model):
    def __init__(self, max_mip=14, name='pyramid', params={}):
        super().__init__()

        self.best_val = 1000000
        self.name = name
        self.params = params

        self.mip_processors = {}
        self.mip_downsamplers = {}
        self.max_mip = max_mip

        for i in range(self.max_mip):
            self.mip_downsamplers[i] = self.default_downsampler()

    def default_downsampler(self):
        return nn.AvgPool2d(2, count_include_pad=False)

    def set_mip_processor(self, module, mip):
        self.mip_processors[mip] = module

    def unset_mip_processor(self, mip):
        del self.mip_processors[mip]

    def set_mip_downsampler(self, module, mip):
        self.mip_downsamplers[mip] = module

    def unset_mip_downsampler(self, mip):
        self.mip_downsamplers[mip] = self.default_downsampler()

    def compute_downsamples(self, img, curr_mip, max_mip):
        downsampled_src_tgt = {}
        downsampled_src_tgt[curr_mip] = img
        #print ("input shape: {}".format(downsampled_src_tgt[curr_mip].shape))
        for mip in range(curr_mip + 1, max_mip + 1):
            # logger.debug("Downsampl MIP {}".format(mip))
            downsampled_src_tgt[mip] = self.mip_downsamplers[mip](
                                                downsampled_src_tgt[mip - 1])
            #print ("downsample {} shape: {}".format(mip, downsampled_src_tgt[mip].shape))
        return downsampled_src_tgt

    def get_all_processor_params(self):
        params = []
        for mip, module in iteritems(self.mip_processors):
            params.extend(module.parameters())
        return params

    def get_processor_params(self, mip):
        params = []
        params.extend(self.mip_processors[mip].parameters())
        return params


    def get_all_downsampler_params(self):
        params = []
        for mip, module in iteritems(self.mip_downsamplers):
            params.extend(module.parameters())
        return params

    def get_downsampler_params(self, mip):
        return [self.mip_downsamplers[mip].parameters()]

    def get_all_params(self):
        params = []
        params.extend(self.get_all_processor_params())
        params.extend(self.get_all_downsampler_params())
        return params

    def upsample_residuals(self, residuals):
        result = nn.functional.interpolate(residuals.permute(
                                         0, 3, 1, 2), scale_factor=2).permute(0, 2, 3, 1)
        result *= 2

        return result

    def get_downsampled_src(self, downsampled_src_tgt, mip):
        emb_dims = downsampled_src_tgt[mip].shape[1]
        src = downsampled_src_tgt[mip][:, 0:emb_dims//2]
        return src

    def get_downsampled_tgt(self, downsampled_src_tgt, mip):
        emb_dims = downsampled_src_tgt[mip].shape[1]
        tgt = downsampled_src_tgt[mip][:, emb_dims//2:]
        return tgt

    def forward(self, src_tgt, mip_in):
        src_tgt_var = Variable(src_tgt.cuda(), requires_grad=True)
        # Find which mips are to be applied
        mips_to_process = iteritems(self.mip_processors)
        filtered_mips   = [(int(m[0]), m[1]) for m in mips_to_process if int(m[0]) >= mip_in]
        ordered_mips    = list(reversed(sorted(filtered_mips)))

        high_mip = ordered_mips[0][0]
        low_mip  = ordered_mips[-1][0]

        # Set up auxilary structures
        residuals  = {}
        downsampled_src_tgt = self.compute_downsamples(src_tgt_var,
                                                       mip_in,
                                                       high_mip)
        # logger.debug("Setting downsample MIP {}".format(high_mip))
        aggregate_res = None
        aggregate_res_mip = None
        # Goal of this loop is to compute aggregate_res
        for mip, module in ordered_mips:
            # generate a warped image at $mip
            if aggregate_res is not None:
                while aggregate_res_mip > mip:
                    aggregate_res = self.upsample_residuals(aggregate_res)
                    aggregate_res_mip -= 1
                if 'rollback' in self.params and self.params['rollback']:
                    tmp_aggregate_res = aggregate_res
                    for i in range(mip_in, aggregate_res_mip):
                        tmp_aggregate_res = self.upsample_residuals(tmp_aggregate_res)

                    raw_src_tgt = downsampled_src_tgt[mip_in] #self.get_downsampled_src(downsampled_src_tgt, mip_in)
                    src_tgt = res_warp_img(
                                       raw_src_tgt,
                                       tmp_aggregate_res)
                    for m in range(mip_in, mip):
                        src_tgt = self.mip_downsamplers[m](src_tgt)
                    src = src_tgt[:, :src_tgt.shape[1]//2]
                else:
                    raw_src = self.get_downsampled_src(downsampled_src_tgt, mip)
                    src = res_warp_img(
                                       raw_src,
                                       aggregate_res)
            else:
                src = self.get_downsampled_src(downsampled_src_tgt, high_mip)

            # Compute residual at level $mip
            tgt = self.get_downsampled_tgt(downsampled_src_tgt, mip)
            sample = torch.cat((src, tgt), 1)
            residuals[mip] = self.mip_processors[mip](sample).permute(0, 2, 3, 1)
            if 'debug' in self.params and self.params['debug']:
                print ("Mip {:2d}: Average {:.2E} Max {:.2E}".format(mip,
                    torch.mean(torch.abs(residuals[mip][..., 4:-4, 4:-4, :])),
                    torch.max(torch.abs(residuals[mip][..., 4:-4, 4:-4, :]))))
            # initialize aggregate flow if None
            if aggregate_res is None:
                aggregate_res = torch.zeros(residuals[mip].shape, device='cuda')
                aggregate_res_mip = mip

            # Add the residual at level $mip to $aggregate_flow
            aggregate_res = combine_residuals(
                                              residuals[mip],
                                              aggregate_res,
                                              is_pix_res=True)

        while aggregate_res_mip > mip_in:
            aggregate_res = self.upsample_residuals(aggregate_res)
            aggregate_res_mip -= 1
        return aggregate_res



from .model import Model
from .scalenet import ScaleNet
from .sequence import Sequence
from .upsample_residuals import UpsampleResiduals
from .tools import create_model_nonportable, dict_to_str

from . import residuals



import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

###############
# Initializer #
###############
class _Initializer(nn.Module):
  def __init__(self, module, per_channel):
    super().__init__()
    self.module = module
    self.per_channel = per_channel
    self.initializing = True

  def initialize(self, x):
    raise NotImplementedError

  def forward(self, x):
    if self.initializing:
      self.initialize(x)
      self.initializing = False

    y = self.module(x)

    return y

  @property
  def weight(self):
    return self.module.weight

  @property
  def bias(self):
    return self.module.bias

  def expand_dims(self, x, ndim):
    for i in range(ndim-1):
      x = x.unsqueeze(-1)
    return x

  def mean(self, x, per_channel=False):
    c = x.shape[1]
    if per_channel:
      return x.transpose(0,1).contiguous().view(c, -1).mean(1).detach()
    else:
      return x.mean().detach()

  def std(self, x, per_channel=False):
    c = x.shape[1]
    if per_channel:
      return x.detach().transpose(0,1).contiguous().view(c, -1).var(1).mean().sqrt()
    else:
      return x.std().detach()

class KyleInitializer(_Initializer):
  def __init__(self, module, per_channel=True, init_bias=True):
    super().__init__(module, per_channel)
    self.init_bias = init_bias

  def initialize(self, x):
    # init weight
    nn.init.kaiming_normal_(self.weight.data, a=0, mode='fan_in')
    y = self.module(x)
    sigma = self.std(y, per_channel=False)
    scale = 1.0 / sigma * np.sqrt(2.0)
    self.weight.data *= self.expand_dims(scale, self.weight.ndimension())

    # init bias
    if self.bias is not None:
      self.bias.data.zero_()
    if self.bias is not None and self.init_bias:
      y = self.module(x)
      mu = self.mean(y, self.per_channel)
      self.bias.data -= mu



import torch
import warnings

class UpsampleResiduals(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, res, scale=2, **kwargs):
        if res.shape[3] == 2 and res.shape[1] == 2:
            warnings.warn("Ambiguous Residual: both 2nd and 4th dimensions of residual are 2.")

        if res.shape[3] != 2 and res.shape[1] == 2:
            channel_permute = False
        else:
            channel_permute = True

        if channel_permute:
            res = res.permute(0, 3, 1, 2)

        result = torch.nn.functional.interpolate(res, scale_factor=scale)
        result *= scale

        if channel_permute:
            result = result.permute(0, 2, 3, 1)

        return result




from torch import nn
from torch.autograd import Variable
import numpy as np
import torch
import six
import os
import random

from .model import Model
from .tools import dict_to_str
from .initialization import KyleInitializer

class Sequence(Model):
    def initc(self, m, mult, kyleinit=False):
        if kyleinit:
           m = KyleInitializer(m, per_channel=True, init_bias=True)
        else:
            m.weight.data *= mult
            m.bias.data[...] = 0

    def validate_arch_desc(self, arch_desc):
        #TODO
        return

    def __init__(self, arch_desc):
        super().__init__()
        self.name = dict_to_str(arch_desc)
        self.params = self.parse_arch_desc(arch_desc)
        self.construct_layers(self.params)

    def parse_arch_desc(self, arch_desc):
        self.validate_arch_desc(arch_desc)
        params = {
            'flags': {
                "use_batchnorm": False,
                "use_inputnorm": False,
                "use_instancenorm": False
            },
            'conv': {
                'k': 7,
                'init_mult': np.sqrt(6),
                'kyleinit': False
            },
            'act': {
                'constructor': nn.LeakyReLU
            },
            'structure': {
                'fms': None,
                'skips': {}
            }
        }

        self.parse_conv_params(params, arch_desc)
        self.parse_act_params(params, arch_desc)
        self.parse_structure_params(params, arch_desc)
        self.parse_flag_params(params, arch_desc)
        return params

    def construct_layers(self, params):
        fms = params['structure']['fms']
        skips = params['structure']['skips']

        init_mult = params['conv']['init_mult']
        kyleinit = params['conv']['kyleinit']
        k = params['conv']['k']
        pad = (k - 1) // 2

        act_constructor = params['act']['constructor']
        self.layers = torch.nn.ModuleList()

        if params['flags']['use_inputnorm']:
            self.layers.append(torch.nn.InstanceNorm2d(num_features=fms[0]))

        for i in range(len(fms) - 1):
            self.layers.append(nn.Conv2d(fms[i], fms[i + 1], k, padding=pad))
            self.initc(self.layers[-1], init_mult, kyleinit=kyleinit)

            if i != len(fms) - 2:
                if params['flags']['use_batchnorm']:
                    self.layers.append(nn.BatchNorm2d(fms[i + 1]))
                if params['flags']['use_instancenorm']:
                    self.layers.append(torch.nn.InstanceNorm2d(
                                                  num_features=fms[i + 1]))
                self.layers.append(act_constructor())
        self.seq = nn.Sequential(*self.layers)

    def parse_conv_params(self, params, arch_desc):
        if 'kyleinit' in arch_desc:
            params['conv']['kyleinit'] = arch_desc['kyleinit']

        if 'k' in arch_desc:
            params['conv']['k'] = arch_desc['k']

        if 'conv_init_mult' in arch_desc:
            params['conv']['init_mult'] = arch_desc['conv_init_mult']

        if 'initc_mult' in arch_desc:
            params['conv']['init_mult'] = arch_desc['initc_mult']

    def parse_act_params(self, params, arch_desc):
        if 'act' in arch_desc:
            act = arch_desc['act']
            if act == 'lrelu':
                params['act']['constructor'] = nn.LeakyReLU
            elif act == 'tanh':
                params['act']['constructor'] = nn.Tanh
            if act == 'relu':
                params['act']['constructor'] = nn.ReLU
            else:
                raise Exception("Unrecognized Activation")

    def parse_flag_params(self, params, arch_desc):
        if 'flags' in arch_desc and 'batchnorm' in arch_desc['flags']:
            params['flags']['use_batchnorm'] = True

        if 'flags' in arch_desc and 'instancenorm' in arch_desc['flags']:
            params['flags']['use_instancenorm'] = True

        if 'flags' in arch_desc and 'inputnorm' in arch_desc['flags']:
            params['flags']['use_inputnorm'] = True

    def parse_structure_params(self, params, arch_desc):
        params['structure']['fms'] = arch_desc['fms']
        if 'skips' in arch_desc:
            params['structure']['skips'] = {
                    int(s): e for (s, e) in six.iteritems(arch_desc['skips'])
            }


    def forward(self, x, *kargs, **kwargs):
        #return self.seq(x)
        skip_data = {}
        count = 0
        skips = self.params['structure']['skips']
        for l in self.layers:
            if isinstance(l, torch.nn.modules.conv.Conv2d):
                count += 1
                if count in skip_data or str(count) in skip_data:
                    #print ('Getting {} from the skip bank'.format(torch.mean(skip_data[count])))
                    #print (torch.mean(torch.abs(x)), torch.var(x))
                    #print (torch.mean(torch.abs(skip_data[count])), torch.var(skip_data[count]))
                    x += skip_data[count]

            if isinstance(l, self.params['act']['constructor']):
                if count in skips or str(count) in skips:
                    #print ('Saving {} to the skip bank'.format(torch.mean(x)))
                    skip_data[skips[count]] = x
            x = l(x)
        return x



import sys
import os

import torch
from torch import nn

class Model(nn.Module):
    def __init__(self):
        super().__init__()
        self.best_val = sys.maxsize
        self.name = "default"
        self.params = {}

    def forward(self, x):
        return x

    def save_state_dict(self, checkpoint_folder='./checkpoints/'):
        path = os.path.join(checkpoint_folder, '{}.state.pth.tar'.format(self.name))
        torch.save(self.state_dict(), path)

    def save_nonportable(self, checkpoint_folder='./checkpoints/'):
        path = os.path.join(checkpoint_folder, '{}.pth.tar'.format(self.name))
        torch.save({'model': self}, path)

    def get_all_params(self):
        result = []
        result.extend(self.parameters())
        return result





import numpy as np
import random
import os
import six
import copy

import torch

from .residuals import res_warp_img, res_warp_res, combine_residuals
from .model import Model
from .add import Add

class ScaleNet(Model):
    def __init__(self, max_level=14, name='pyramid', params={}):
        super().__init__()

        self.name = name
        self.params = params
        self.max_level = max_level

        self.level_upmodules = torch.nn.ModuleDict()
        self.level_downmodules = torch.nn.ModuleDict()
        self.level_uplinks = torch.nn.ModuleDict()
        self.level_downlinks = torch.nn.ModuleDict()
        self.level_skiplinks = torch.nn.ModuleDict()
        self.level_combiners = torch.nn.ModuleDict()

        for i in range(self.max_level):
            self.level_uplinks[str(i)]   = self.default_uplink()
            self.level_downlinks[str(i)] = self.default_downlink()
            self.level_skiplinks[str(i)] = self.default_skiplink()
            self.level_combiners[str(i)] = self.default_combiner()

    def forward(self, x, level_in=0, multimip_input=False):
        state = {}
        state = self.up_path(x, level_in, multimip_input, state)
        result = self.down_path(state)
        self.state = state
        return result


    def up_path(self, x_in, level_in, multimip_input, state):
        state['up'] = {}
        all_module_levels = list(self.level_upmodules.keys()) + list(self.level_downmodules.keys())
        max_level = max([int(i) for i in all_module_levels])

        x = None
        for level in range(level_in, max_level + 1):
            level_state = {}
            state['up'][str(level)] = level_state

            if multimip_input and level in x_in:
                x = x_in[level]
            elif x is None:
                assert isinstance(x_in, torch.Tensor)
                x = x_in


            level_state['input'] = x

            if str(level) in self.level_upmodules:
                x = self.level_upmodules[str(level)](x, state, level)

            level_state['output'] = x

            skip = self.level_skiplinks[str(level)](x)
            level_state['skip'] = skip

            x = self.level_uplinks[str(level)](x)
            level_state['uplink'] = x

        return state

    def down_path(self, state):
        state['down'] = {}
        prev_out = None
        max_level = max([int(i) for i in state['up'].keys()])
        min_level = min([int(i) for i in state['up'].keys()])
        for level in reversed(range(min_level, max_level + 1)):
            level_state = {}
            state['down'][str(level)] = level_state

            skip = state['up'][str(level)]['skip']

            if prev_out is None:
                level_in = torch.zeros_like(skip, device='cuda')
            else:
                downlink = self.level_downlinks[str(level)]
                level_in = downlink(prev_out)

            level_state['input'] = level_in

            if str(level) in self.level_downmodules:
                downmodule_in = self.level_combiners[str(level)](skip, level_in, state, level)
                level_out = self.level_downmodules[str(level)](downmodule_in, state, level)
                if 'debug' in self.params and self.params['debug']:
                    print ("{}: {}".format(level, torch.mean(torch.abs(level_out))))
            else:
                level_out = level_in

            level_state['output'] = level_out
            prev_out = level_out


        result = state['down'][str(min_level)]['output']
        return result

    ############# Defaults ###############

    def default_uplink(self):
        return torch.nn.AvgPool2d(2, count_include_pad=False)

    def default_downlink(self):
        return torch.nn.Upsample(scale_factor=2, mode='bilinear')

    def default_combiner(self):
        return Add()

    def default_skiplink(self):
        return torch.nn.Sequential() #aka identity

    ############# Setters ###############

    def set_upmodule(self, module, level):
        self.level_upmodules[str(level)] = module

    def set_downmodule(self, module, level):
        self.level_downmodules[str(level)] = module

    def set_uplink(self, link, level):
        self.level_uplinks[str(level)] = link
        if len(list(link.parameters())) > 0:
            print ("Uplink mean weight [0] == {}".format(torch.mean(list(link.parameters())[0])))

    def set_downlink(self, link, level):
        self.level_downlinks[str(level)] = link

    def set_skiplink(self, link, level):
        self.level_skiplinks[str(level)] = link

    def set_combiner(self, combiner, level):
        self.level_combiners[str(level)] = combiner

    def set_all_uplinks(self, link):
        for level in range(self.max_level):
            self.level_uplinks[str(level)] = copy.deepcopy(link)

    def set_all_downlinks(self, link):
        for level in range(self.max_level):
            self.level_downlinks[str(level)] = copy.deepcopy(link)

    def set_all_skiplinks(self, link):
        for level in range(self.max_level):
            self.level_skiplinks[str(level)] = copy.deepcopy(link)

    def set_all_combiners(self, combiner):
        for level in range(self.max_level):
            self.level_combiners[str(level)] = copy.deepcopy(combiner)


    ############# Unsetters ###############
    def unset_upmodule(self, level):
        del self.level_upmodules[level]

    def unset_downmodule(self, level):
        del self.level_downmodules[level]

    def unset_uplink(self, level):
        del self.level_uplinks[level]

    def unset_downlink(self, level):
        del self.level_downlinks[level]

    def unset_skiplink(self, level):
        del self.level_skiplinks[level]

    def unset_combiner(self, level):
        del self.level_combiners[level]




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scalenet",
    version="0.0.2",
    author="Sergiy Popovych",
    author_email="sergiy.popovich@gmail.com",
    description="A universal pytorch module for implementing multiscale architectures (UNet, spatial pyramid, SpyNet, res-Unet)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/supersergiy/scalenet",
    packages=setuptools.find_packages(),
)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import numpy as np
import random
from six import iteritems
import os

from .residuals import res_warp_img, res_warp_res, combine_residuals
from .model import Model

class ScaleNet(Model):
    def __init__(self, max_mip=14, name='pyramid', params={}):
        super().__init__()

        self.name = name
        self.params = params
        self.max_mip = max_mip

        self.mip_upmodules = torch.nn.ModuleDict()
        self.mip_downmodules = torch.nn.ModuleDict()
        self.mip_uplinks = torch.nn.ModuleDict()
        self.mip_downlinks = torch.nn.ModuleDict()
        self.mip_skiplinks = torch.nn.ModuleDict()
        self.mip_combiners = torch.nn.ModuleDict()

        for i in range(self.max_mip):
            self.mip_uplinks[str(i)]   = self.default_uplink()
            self.mip_downlinks[str(i)] = self.default_downlink()
            self.mip_skiplinks[str(i)] = self.default_skiplink()
            self.mip_combiners[str(i)] = self.default_combiner()

    def forward(self, x, mip_in):
        up_result = self.up_path(x, mip_in)
        result = self.down_path(up_result)
        return result


    def up_path(self, x, mip_in):
        result = {}
        all_module_mips = list(self.mip_upmodules.keys()) + list(self.mip_downmodules.keys())
        max_mip = max([int(i) for i in all_module_mips])

        for mip in range(mip_in, max_mip + 1):
            result[str(mip)] = {}
            result[str(mip)]['input'] = x
            if str(mip) in self.mip_upmodules:
                x = self.mip_upmodules[str(mip)](x)

            result[str(mip)]['output'] = x

            skip = self.mip_skiplinks[str(mip)](x)
            result[str(mip)]['skip'] = skip

            x = self.mip_uplinks[str(mip)](x)
            result[str(mip)]['uplink'] = x

        return result

    def down_path(self, up_result):
        prev_out = None
        max_mip = max([int(i) for i in up_result.keys()])
        min_mip = min([int(i) for i in up_result.keys()])

        for mip in reversed(range(min_mip, max_mip + 1)):
            skip = up_result[str(mip)]['skip']
            if prev_out is None:
                prev_out = torch.zeros_like(skip, device=skip.get_device())
            if str(mip) in self.mip_downmodules:
                downmodule_in = self.mip_combiners[str(mip)](skip, prev_out)
                out = self.mip_downmodules[str(mip)](downmodule_in)
            else:
                out = prev_out


            prev_out = out

            if mip != min_mip:
                prev_out = self.mip_downlinks[str(mip)](out)

        return prev_out

    def compute_downsamples(self, img, curr_mip, max_mip):
        downsampled_src_tgt = {}
        downsampled_src_tgt[curr_mip] = img
        #print ("input shape: {}".format(downsampled_src_tgt[curr_mip].shape))
        for mip in range(curr_mip + 1, max_mip + 1):
            # logger.debug("Downsampl MIP {}".format(mip))
            downsampled_src_tgt[mip] = self.mip_downsamplers[mip](
                                                downsampled_src_tgt[mip - 1])
            #print ("downsample {} shape: {}".format(mip, downsampled_src_tgt[mip].shape))
        return downsampled_src_tgt


    def get_downsampled_src(self, downsampled_src_tgt, mip):
        emb_dims = downsampled_src_tgt[mip].shape[1]
        src = downsampled_src_tgt[mip][:, 0:emb_dims//2]
        return src

    def get_downsampled_tgt(self, downsampled_src_tgt, mip):
        emb_dims = downsampled_src_tgt[mip].shape[1]
        tgt = downsampled_src_tgt[mip][:, emb_dims//2:]
        return tgt

    '''def forward(self, src_tgt, mip_in):
        src_tgt_var = Variable(src_tgt.cuda(), requires_grad=True)
        # Find which mips are to be applied
        mips_to_process = iteritems(self.mip_processors)
        filtered_mips   = [(int(m[0]), m[1]) for m in mips_to_process if int(m[0]) >= mip_in]
        ordered_mips    = list(reversed(sorted(filtered_mips)))

        high_mip = ordered_mips[0][0]
        low_mip  = ordered_mips[-1][0]

        # Set up auxilary structures
        residuals  = {}
        downsampled_src_tgt = self.compute_downsamples(src_tgt_var,
                                                       mip_in,
                                                       high_mip)
        # logger.debug("Setting downsample MIP {}".format(high_mip))
        aggregate_res = None
        aggregate_res_mip = None
        # Goal of this loop is to compute aggregate_res
        for mip, module in ordered_mips:
            # generate a warped image at $mip
            if aggregate_res is not None:
                while aggregate_res_mip > mip:
                    aggregate_res = self.upsample_residuals(aggregate_res)
                    aggregate_res_mip -= 1
                if 'rollback' in self.params and self.params['rollback']:
                    tmp_aggregate_res = aggregate_res
                    for i in range(mip_in, aggregate_res_mip):
                        tmp_aggregate_res = self.upsample_residuals(tmp_aggregate_res)

                    raw_src_tgt = downsampled_src_tgt[mip_in] #self.get_downsampled_src(downsampled_src_tgt, mip_in)
                    src_tgt = res_warp_img(
                                       raw_src_tgt,
                                       tmp_aggregate_res)
                    for m in range(mip_in, mip):
                        src_tgt = self.mip_downsamplers[m](src_tgt)
                    src = src_tgt[:, :src_tgt.shape[1]//2]
                else:
                    raw_src = self.get_downsampled_src(downsampled_src_tgt, mip)
                    src = res_warp_img(
                                       raw_src,
                                       aggregate_res)
            else:
                src = self.get_downsampled_src(downsampled_src_tgt, high_mip)

            # Compute residual at level $mip
            tgt = self.get_downsampled_tgt(downsampled_src_tgt, mip)
            sample = torch.cat((src, tgt), 1)
            residuals[mip] = self.mip_processors[mip](sample).permute(0, 2, 3, 1)
            if 'debug' in self.params and self.params['debug']:
                print ("Mip {:2d}: Average {:.2E} Max {:.2E}".format(mip,
                    torch.mean(torch.abs(residuals[mip][..., 4:-4, 4:-4, :])),
                    torch.max(torch.abs(residuals[mip][..., 4:-4, 4:-4, :]))))
            # initialize aggregate flow if None
            if aggregate_res is None:
                aggregate_res = torch.zeros(residuals[mip].shape, device='cuda')
                aggregate_res_mip = mip

            # Add the residual at level $mip to $aggregate_flow
            aggregate_res = combine_residuals(
                                              residuals[mip],
                                              aggregate_res,
                                              is_pix_res=True)

        while aggregate_res_mip > mip_in:
            aggregate_res = self.upsample_residuals(aggregate_res)
            aggregate_res_mip -= 1
        return aggregate_res'''
    ############# Defaults ###############

    def default_uplink(self):
        return torch.nn.AvgPool2d(2, count_include_pad=False)

    def default_downlink(self):
        return torch.nn.Upsample(scale_factor=2, mode='bilinear')

    def default_combiner(self):
        return AddModule()

    def default_skiplink(self):
        return nn.Sequential() #aka identity

    ############# Setters ###############

    def set_upmodule(self, module, mip):
        self.mip_upmodules[str(mip)] = module

    def set_downmodule(self, module, mip):
        self.mip_downmodules[str(mip)] = module

    def set_uplink(self, link, mip):
        self.mip_uplinks[str(mip)] = link

    def set_downlink(self, link, mip):
        self.mip_downlinks[str(mip)] = link

    def set_skiplink(self, link, mip):
        self.mip_skiplinks[str(mip)] = link

    def set_combiner(self, combiner, mip):
        self.mip_combiners[str(mip)] = combiner

    ############# Unsetters ###############
    def unset_upmodule(self, mip):
        del self.mip_upmodules[mip]

    def unset_downmodule(self, mip):
        del self.mip_downmodules[mip]

    def unset_uplink(self, mip):
        del self.mip_uplinks[mip]

    def unset_downlink(self, mip):
        del self.mip_downlinks[mip]

    def unset_skiplink(self, mip):
        del self.mip_skiplinks[mip]

    def unset_combiner(self, mip):
        del self.mip_combiners[mip]



class AddModule(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x, y):
        return x + y



import torch

def res_warp_res(res_a, res_b, is_pix_res=True):
    if is_pix_res:
        res_b = 2 * res_b / (res_b.shape[-2])

    if len(res_a.shape) == 4:
        result = gridsample_residual(
                        res_a.permute(0, 3, 1, 2),
                        res_b,
                        padding_mode='border').permute(0, 2, 3, 1)
    elif len(res_a.shape) == 3:
        result = gridsample_residual(
                        res_a.permute(2, 0, 1).unsqueeze(0),
                        res_b.unsqueeze(0),
                        padding_mode='border')[0].permute(1, 2, 0)
    else:
        raise Exception("Residual warping requires BxHxWx2 or HxWx2 format.")

    return result


def res_warp_img(img, res_in, is_pix_res=True):

    if is_pix_res:
        res = 2 * res_in / (img.shape[-1])
    else:
        res = res_in

    if len(img.shape) == 4:
        result = gridsample_residual(img, res, padding_mode='zeros')
    elif len(img.shape) == 3:
        result = gridsample_residual(img.unsqueeze(0),
                                     res.unsqueeze(0), padding_mode='zeros')[0]
    elif len(img.shape) == 2:
        result = gridsample_residual(img.unsqueeze(0).unsqueeze(0),
                                     res.unsqueeze(0),
                                     padding_mode='zeros')[0, 0]
    else:
        raise Exception("Image warping requires BxCxHxW or CxHxW format." +
                        "Recieved dimensions: {}".format(len(img.shape)))

    return result


def combine_residuals(a, b, is_pix_res=True):
    return b + res_warp_res(a, b, is_pix_res=is_pix_res)

upsampler = torch.nn.UpsamplingBilinear2d(scale_factor=2)
def upsample_residuals(residuals):
    result = upsampler(residuals.permute(
                                     0, 3, 1, 2)).permute(0, 2, 3, 1)
    result *= 2
    return result

def gridsample(source, field, padding_mode):
    if source.shape[2] != source.shape[3]:
        raise NotImplementedError('Grid sample is not impolemented for non-square tensors.')
    scaled_field = field * source.shape[2] / (source.shape[2] - 1)
    return torch.nn.functional.grid_sample(source, scaled_field, mode="bilinear",
                                            padding_mode=padding_mode)

def get_identity_grid(size):
    with torch.no_grad():
        id_theta = torch.cuda.FloatTensor([[[1,0,0],[0,1,0]]]) # identity affine transform
        I = torch.nn.functional.affine_grid(id_theta,torch.Size((1,1,size,size)))
        I *= (size - 1) / size # rescale the identity provided by PyTorch
        return I

def gridsample_residual(source, res, padding_mode):
    size = source.size()[-1]
    field = res + get_identity_grid(size)
    return gridsample(source, field, padding_mode)



import os
import torch

def dict_to_str(d, separator='_'):
    result = ''
    if isinstance(d, dict):
        for k in sorted(d.keys()):
            if result != '':
                result += separator
            result += 'start' + separator + k
            result += separator
            result += dict_to_str(d[k])
            result += separator
            result += 'end' + separator + k
    else:
        result = str(d)
        if separator in result:
            raise ValueError('Separator ({}) cannot occur within transformed dictionary.\
                              Violation: {}'.format(separator, result))
        if d in ['start', 'end']:
            raise ValueError('Keywords "start" and "end" cannot occur within transformed dictionary')

    return result

def upsample_residuals(residuals):
    result = nn.functional.interpolate(residuals.permute(
                                     0, 3, 1, 2), scale_factor=2).permute(0, 2, 3, 1)
    result *= 2

    return result

def create_model_nonportable(model_constructor, constructor_params={},
                             name=None, source_model_path=None,
                             checkpoint_folder='./checkpoints/', cuda=True):
    model_path = os.path.join(checkpoint_folder, "{}.pth.tar".format(name))
    if (name is not None) and (os.path.isfile(model_path)):
        print("Loading from checkpoint")
        model = torch.load(model_path)['model']
    elif source_model_path is not None:
        print("Loading weights from {}".format(init_path))
        model = torch.load(source_model_path)['model']
        model.best_val = sys.maxsize
    else:
        print("Initializing new model {}".format(name))
        model = model_constructor(**constructor_params)

    model.name = name
    if cuda:
        return model.cuda()
    else:
        return model
